/* jmixer.cpp - part of MUSE
 * (c)2000 Denis Roio aka jaromil
 * c++ engine for multiple mp3 realtime stream mixing
 */

#include <iostream.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <termios.h>
#include <sys/time.h>

#include "jmixer.h"
#include "audioproc.h"

Stream_mixer::Stream_mixer() {

  dsp = 0;
  have_gui = false;

  fprintf(stderr,"Initializing channels ...");

#ifdef MPG123
  
  /* init channels */
  for(c=0;c<MAX_CHANNELS;c++) {
    mp3ch[c].dsp = &dsp;
    mp3ch[c].sample_rate = SAMPLE_RATE;
  }

#endif

  fprintf(stderr," done.\n");

  quit = false;

  /* Initialization as a synchronous object */
  synchronous.registerObjectName("Stream_mixer");

}

void Stream_mixer::run() {

  if(debug)
    cerr << selfName() << " running on thread " << (int)pthread_self() << "\n";

  while(!quit) {
    
    synchronous.select.accept(0, id_stream_mixer);

    synchronous.select.waitUntil(1, LTI::Kernel::now()+5);
    synchronous.wait();

    cafudda();
  }

  cerr << selfName() << " exiting " << "\n";

}

Stream_mixer::~Stream_mixer() {

#ifdef MPG123
  for(i=0;i<MAX_CHANNELS;i++)
    mp3ch[i].stop();
#endif

  if(dsp!=0)
    close_soundcard();
  
  if(icecastout)
    stop_icecast();

}

void Stream_mixer::register_gui(Amuse *reg_gui) { gui = reg_gui; have_gui = true; }

bool Stream_mixer::open_soundcard()
{
    int format,tstereo,speed,caps;
    static int val;
    int hops, frag;

    if((dsp=open("/dev/dsp",O_RDWR))==-1) {
        perror("Error opening /dev/dsp: ");
	return(false);
    } else
      fprintf(stderr,"Found soundcard on /dev/dsp, initializing ...\n");

    /* let's use NE - it works also with LE */
    format = AFMT_S16_NE;

    tstereo = 1; /* i implement only stereo dsp output 'til anybody complains */
    speed = SAMPLE_RATE; /* acc... */

    ioctl(dsp, SNDCTL_DSP_GETCAPS, &caps);
    if(caps & DSP_CAP_DUPLEX) {
      fullduplex = true;
      fprintf(stderr,"full duplex supported. good.\n");
      ioctl(dsp, SNDCTL_DSP_SETDUPLEX, 0);
      ioctl(dsp, DSP_CAP_DUPLEX, 0);
    } else {
      fprintf(stderr,"full duplex is not supported.\n");
      fullduplex = false;
    }

     /* BUFFER FRAGMENTATION */
    frag = FRAGSIZE;
    for (hops=0; frag >>=1; hops++);	
    val = (FRAGCOUNT << 16) + hops;
    if(ioctl(dsp,SNDCTL_DSP_SETFRAGMENT,&val)==-1)
      fprintf(stderr,"failed to set dsp buffers\n");

    /* FORMAT */
    if(ioctl(dsp,SNDCTL_DSP_SETFMT,&format)==-1)
      fprintf(stderr,"failed to set data format\n");

    /* CHANNELS */
    if(ioctl(dsp,SNDCTL_DSP_STEREO,&tstereo)==-1)
      fprintf(stderr,"something went wrong with the stereo setting\n");

    /* SAMPLERATE */
    if(ioctl(dsp,SNDCTL_DSP_SPEED,&speed)==-1)
      fprintf(stderr,"speed setting failed\n");

    /* GET FRAG SIZE BACK */
    if(ioctl(dsp,SNDCTL_DSP_GETBLKSIZE,&val)==-1)
      fprintf(stderr,"get block size failed to return\n");

    fprintf(stderr,"dsp setted to 16bit %dHz stereo\n",SAMPLE_RATE);

    if(fullduplex) dspout = true; /* which should be the default? */

    livein = new LiveIn(SAMPLE_RATE, 2, &dsp);

    return(true);
} /* open_soundcard */

bool Stream_mixer::set_icecast(char *ip, int port, char *pass, char *mountpoint, int bps) {

  ice.set(ip,port,pass,mountpoint,bps);
  ice.init_encoder(bps);
  
  if(ice.start()) {
    icecastout = true;
    fprintf(stderr,"Estabilished %ubps stream on %s:%u/%s\n",bps,ip,port,mountpoint);
  } else {
    icecastout = false;
    fprintf(stderr,"Got problems streaming on server %s. No milk??\n",ip);
  }

  synchronous.reply();
  return(icecastout);
}

void Stream_mixer::stop_icecast() {
  ice.stop();
  icecastout = false;

  fprintf(stderr,"Stopped streaming to icecast server\n");

  synchronous.reply();
}

void Stream_mixer::cafudda()
{
  int i,j,num;
  float res;

  j=BUF_SIZE>>1;
  
  memset(process_buffer,0,j<<2);

#ifdef MPG123
  
  for(i=0;i<MAX_CHANNELS;i++) {
    if((mp3ch[i].pid)&&(mp3ch[i].on)) {
      mp3ch[i].get_audio();
      if(mp3ch[i].volume!=1.0)
	for(j=0;j<num;j++)
	  process_buffer[j] = (int) (process_buffer[j] + (mp3ch[i].gotin[j] * mp3ch[i].volume));
      else
	for(j=0;j<num;j++)
	  process_buffer[j] = (int) (process_buffer[j] + mp3ch[i].gotin[j]);

      /* if the mp3 reached the end */
      if(mp3ch[i].end)
	mp3ch[i].ends();

    } // if (mp3ch[i].pid&&mp3ch[i].on)
  } // for i=0 to MAX_CHANNELS

#else

  /* math formula for rounding up floats:
     int(x*10^n+0.5)/10^n
     why here? just a reminder...
  */

  for(i=0;i<MAX_CHANNELS;i++) {
    if(xach[i].on) {
      res = xach[i].get_audio();

      /* let's notice the new position */
      if(have_gui) 
	upd_timeline(i,res);

    }
    if(xach[i].on) {
      if(xach[i].buffer->channels == 1) /* mono */
	for(num=j;num>0;num--)
	  process_buffer[num] = (int) (process_buffer[num] + ((((int)xach[i].buffer->pcm_samples[num])<<8)&(xach[i].buffer->pcm_samples[num]))*xach[i].volume);
      else /* stereo */
	for(num=j;num>0;num--)
	  process_buffer[num] = (int) (process_buffer[num] + (((int)xach[i].buffer->pcm_samples[num])*xach[i].volume));

      
    } /* if(xach[i].on) */
  } /* for(i=0;i<MAX_CHANNELS;i++) */
#endif
  
  if(livein->on) {
    livein->get_audio();
    for(num=0;num<j;num++)
      process_buffer[num] = (int) (process_buffer[num] + livein->gotin[num]);
  }
  
  /* CLIPPING */
  clip_audio(process_buffer,audio_buffer,j);
  /* 00000000 */

  /* get level */
  fprintf(stderr,"peak on %u\n",get_level(audio_buffer,j<<1));


  /* SEND 2 ICECAST */
  if(icecastout) {
    num = ice.encode(&audio_buffer[0],j>>1,process_buffer,BUF_SIZE);
    if(num>0) {
      ice.sync();
      if(!ice.send((short int *)process_buffer,num))
	fprintf(stderr,"Send error: cannot send stream to icecast server\n");
      else if(num==-1) {
	if(have_gui) {
	  fl_alert("Encoding failed: mp3buffer overflowed\nUACK!! bailing out!");
	  exit(-1);
	} else {
	  fprintf(stderr,"Encoding failed: mp3buffer overflowed\nUACK!! bailing out!\n");
	  exit(-1);
	}
      } 
    }
  }
  /* 00000000000000 */

  /* WRITE 2 DSP */
  if(dspout)
    write(dsp,audio_buffer,j<<1);
  /* 00000000000 */
}

void Stream_mixer::pause_channel(int ch) {

#ifdef MPG123
  mp3ch[ch].on = !mp3ch[ch].on;
#else
  if(xach[ch].opened)
    xach[ch].on = !xach[ch].on;
#endif

  synchronous.reply();
}

bool Stream_mixer::set_channel(int ch, int playlist_pos) {

  bool res;

#ifdef MPG123
  mp3ch[ch].start(playlist_pos);
#else

  res = xach[ch].set(playlist_pos);
  if(!res) {
    if(have_gui) {
      char temp[MAX_PATH_SIZE+100];
      sprintf(temp,"It seems there's something missing. I tryed to open %s, check that out.\n",xach[ch].playlist[playlist_pos]);
      fl_alert(temp);
    } else {
      fprintf(stderr,"error opening file %s\n", xach[ch].playlist[playlist_pos]);
    }
  } else 
#endif

  synchronous.reply();
  return(res);
}

void Stream_mixer::set_volume(int ch, float vol) {

#ifdef MPG123
  mp3ch[ch].volume = vol;
#else
  xach[ch].volume = vol;
#endif

  synchronous.reply();
}

void Stream_mixer::set_speed(int ch, float speed) {

#ifdef MPG123
  mp3ch[ch].speed = speed;
#endif

  synchronous.reply();
}

bool Stream_mixer::set_playlist(int ch, char *pl) {

  FILE *plist;
  int c = 0;
  char *p;

#ifdef MPG123
  mp3ch[ch].get_playlist(pl);
#else

  xach[ch].get_playlist(pl);

  plist = fopen(pl,"r");
  if(plist == NULL) {
    fprintf(stderr,"I can't read the playlist file!\n");
    return(false);
  }

  while (fgets(&xach[ch].playlist[c][0],MAX_PATH_SIZE,plist) != NULL) {
    p = &xach[ch].playlist[c][0];
    while(*p!='\n') p++;
    *p = '\0';
    if(have_gui) {
      while(*p!='/') p--; /* gives to the gui the filename without path */
      gui->chan[ch]->playlist_browser->add(p+1,NULL);
    }
    c++;
  }

  fclose(plist);
  
  xach[ch].playlist_len = c;
  
#endif

  synchronous.reply();
  return(true);
}

void Stream_mixer::play_channel(int ch) {

#ifdef MPG123
  mp3ch[ch].on = true;
#else
  xach[ch].play();
#endif

  synchronous.reply();
}
  
bool Stream_mixer::stop_channel(int ch) {

  bool res = true;

  if(ch<=MAX_CHANNELS) {

#ifdef MPG123
    mp3ch[ch].stop();
#else
    res = xach[ch].stop();
#endif
  }

  synchronous.reply();
  return(res);

}

#ifndef MPG123
bool Stream_mixer::set_position(float pos, int ch) {
  bool res = true;

  res = xach[ch].pos(pos);

  synchronous.reply();
  return(res);
}
#endif

void Stream_mixer::dump_status() {

  fprintf(stderr,"__________dumping_status__________\n");

#ifdef MPG123
  for(c=0;c<MAX_CHANNELS;c++) {
    
    if(!mp3ch[c].on)
      fprintf(stderr,"[channel %02u OFF] : PID %u\n",c,mp3ch[c].pid);
    else {
      fprintf(stderr,"[channel %02u ON ] : PID %u\n",c,mp3ch[c].pid);
      fprintf(stderr,"\tname:\t%s\n",mp3ch[c].name);
      fprintf(stderr,"\tvolume:\t%f\n",mp3ch[c].volume);
      fprintf(stderr,"\tspeed:\t%f\n",mp3ch[c].speed);
      fprintf(stderr,"\trate:\t%u\n",mp3ch[c].rate);
      fprintf(stderr,"\tchannels:\t%u\n",mp3ch[c].channels);
      fprintf(stderr,"\tend:\t%u\n",mp3ch[c].end);
      fprintf(stderr,"\tP stream:\t%u\n",mp3ch[c].stream[0]);
      fprintf(stderr,"\tC stream:\t%u\n",mp3ch[c].stream[1]);
    }
  }
#endif

  fprintf(stderr,"__________________________________\n");    
  synchronous.reply();
}

bool Stream_mixer::set_live(bool stat) {
  if(!((dspout)&&(!fullduplex)&&(stat)))
    livein->on = stat;
  synchronous.reply();
  return(livein->on&stat);
}

void Stream_mixer::close_soundcard() {
  ioctl(dsp, SNDCTL_DSP_RESET, 0);
  close(dsp);
}

void Stream_mixer::upd_timeline(int ch, float res) {
  char temp[10];

  if((res - xach[ch].position)>0.01) {
    gui->chan[ch]->timeline_slider->value(res);
    xach[ch].position = res;
  }

  sprintf(temp,"%02u:%02u",xach[ch].status->timecode.m,xach[ch].status->timecode.s);
  gui->chan[ch]->time->value(temp);

}
