/* outchannels.cpp - MuSE
   (c)2000 Denis Roio aka jaromil
   see COPYING for licensing stuff

   different classes for different OUT channels
   they are instantiated and used by the Stream_mixer class (jmixer.cpp)
*/

#include <iostream.h>
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "generic.h"
#include "outchannels.h"

IcecastOut::IcecastOut() {

  initialized = false;
  running = false;

}

IcecastOut::~IcecastOut() {
  if(running)
    stop();
}

bool IcecastOut::start() {

  if(initialized&&!running) {
    shout_init_connection(&ice_conn);
    ice_conn.ip = ip;
    ice_conn.port = port;
    ice_conn.password = pass;
    ice_conn.mount = mountpoint;
    ice_conn.bitrate = bps*1024;
    
    fprintf(stderr,"Contacting icecast server ...");
    
    if(shout_connect(&ice_conn)) {
      fprintf(stderr," done\n");
      running = true;
    } else {
      fprintf(stderr," ops!\nunable to connect or reach host\n");
      running = false;
    }
    return(running);
  } else
    return(false); /* icecast connection not initialized or already running */
}

void IcecastOut::stop() {
  if(running) {
    shout_disconnect(&ice_conn);
    running = false;
  }
}

bool IcecastOut::set(char *naddress, int nport, char *npass, char *nmountpoint, int nbps) {
  struct sockaddr_in server_addr;
  struct hostent *h;

  if(running)
    stop();

  memset(&server_addr, 0, sizeof(struct sockaddr_in));
  server_addr.sin_family = AF_INET;

#ifdef HAVE_INET_ATON
  if (!(inet_aton(naddress,&server_addr.sin_addr))) {
#else
  if ((server_addr.sin_addr.s_addr=inet_addr(naddress)) < 0) {
#endif

    if(!(h = gethostbyname(naddress))) {
      printf("Unable to reach icecast server %s\n",naddress);
      return(false);
    }

    server_addr.sin_family = h->h_addrtype;
    memcpy(&server_addr.sin_addr, h->h_addr, h->h_length);
  }

  sprintf(ip,"%s",inet_ntoa(*(struct in_addr *) &server_addr.sin_addr.s_addr));

  port = nport;
  strncpy(pass,npass,64);
  strncpy(mountpoint,nmountpoint,64);
  bps = nbps;

  initialized = true;
  return(initialized);
}

bool IcecastOut::send(short *sendbuf, long len) {
  if(running) {
    int res;
    res = shout_send_data(&ice_conn, (unsigned char *)sendbuf, len);
    if(!res) {
      fprintf(stderr,"Send error: %i\n",ice_conn.error);
      return(false);
    }
    return(res);
  } else
    return(false);
}

void IcecastOut::sync() {
  shout_sleep(&ice_conn);
}

void IcecastOut::init_encoder(int bps) {
  
  fprintf(stderr,"Initializing mp3 encoder engine to %u ...",bps);

  lame_init(&enc_flags);
  enc_flags.num_channels = 1; /* the coming out stream is hardcoded stereo */
  enc_flags.in_samplerate = SAMPLE_RATE;
  enc_flags.brate = bps;
  enc_flags.mode = 1; /* jstereo */
  enc_flags.quality = QUALITY;
  enc_flags.error_protection = 1; /* use 2 bytes per frame for a CRC checksum */

  /* lame chooses for us frequency filtering */
  enc_flags.lowpassfreq = 0;
  enc_flags.highpassfreq = 0;

  lame_init_params(&enc_flags);

  fprintf(stderr," done\n");
}

int IcecastOut::encode(short int *pcm,int num,int *dest,int size) {
  int res;
  
  res = lame_encode_buffer_interleaved(&enc_flags, pcm, num, (char *) dest, size);
  return(res);
}
